from django.test import override_settings, TestCase
from django.urls import reverse

from parameterized import parameterized

from power_query.defaults import create_defaults
from power_query.fixtures import (
    FormatterFactory,
    QueryFactory,
    ReportFactory,
    UserFactory,
)
from power_query.models import Report

CONTENT_TYPES = [
    ("application/json", "application/json"),
    ("text/html", "text/html; charset=utf-8"),
    ("text/plain", "text/plain"),
]


@override_settings(POWER_QUERY_DB_ALIAS="default")
class TestPowerQueryResponses(TestCase):
    databases = {"default"}

    @classmethod
    def setUpTestData(cls) -> None:
        cls.superuser = UserFactory(is_superuser=True, is_staff=True, is_active=True)
        cls.user1 = UserFactory(is_superuser=False, is_staff=False, is_active=True)
        cls.user2 = UserFactory(is_superuser=False, is_staff=False, is_active=True)
        create_defaults()
        cls.formatter_json = FormatterFactory(
            name="Queryset To JSON", content_type="json", code=""
        )
        cls.query = QueryFactory()
        cls.report1: Report = ReportFactory(
            formatter=cls.formatter_json, query=cls.query, owner=cls.user1
        )
        cls.report2: Report = ReportFactory(
            formatter=cls.formatter_json, query=cls.query, owner=cls.user2
        )
        cls.report2.execute(run_query=True)

    @parameterized.expand(CONTENT_TYPES)
    def test_fetch_no_auth_content_types(self, accept: str, content_type: str) -> None:
        with self.settings(POWER_QUERY_DB_ALIAS="default"):
            url = reverse("power_query:data", args=[self.report2.pk])
            response = self.client.get(url, HTTP_ACCEPT=accept)
            self.assertEqual(response.status_code, 401)



# Generated by Django 4.2.1 on 2023-07-18 14:47

from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("power_query", "0001_migration_squashed_0014_migration"),
    ]

    operations = [
        migrations.AlterField(
            model_name="dataset",
            name="description",
            field=models.CharField(max_length=100),
        ),
        migrations.AlterField(
            model_name="report",
            name="limit_access_to",
            field=models.ManyToManyField(
                blank=True, related_name="+", to=settings.AUTH_USER_MODEL
            ),
        ),
        migrations.AlterField(
            model_name="reportdocument",
            name="limit_access_to",
            field=models.ManyToManyField(
                blank=True, related_name="+", to=settings.AUTH_USER_MODEL
            ),
        ),
    ]



import json
from typing import Any

from django import template
from django.utils.safestring import mark_safe

from adminactions.utils import get_attr

from power_query.utils import get_sentry_url, sizeof

register = template.Library()


@register.filter()
def field(obj: Any, field_name: str) -> str:
    return get_attr(obj, field_name)


@register.filter()
def link_to_sentry(event_id: Any, href: bool = False) -> str:
    return get_sentry_url(event_id, href)


@register.filter(name="classname")
def get_class(value: Any) -> Any:
    return value.__class__.__name__


@register.filter()
def dataset_to_json(value: Any) -> str:
    return json.dump(value)  # type: ignore # FIXME: json.dump needs a second argument (fp)


@register.filter()
def fmt_size(value: Any) -> str:
    return mark_safe(sizeof(value))



import logging
from typing import Any, Dict, Optional, Sequence, TYPE_CHECKING

from django import forms
from django.conf import settings
from django.contrib import messages
from django.contrib.admin import ModelAdmin, register
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.db.models import QuerySet
from django.http import HttpRequest, HttpResponse, HttpResponseRedirect
from django.shortcuts import render
from django.template.response import TemplateResponse
from django.urls import reverse

import tablib
from admin_extra_buttons.decorators import button
from admin_extra_buttons.mixins import ExtraButtonsMixin
from adminactions.helpers import AdminActionPermMixin
from adminfilters.autocomplete import AutoCompleteFilter
from adminfilters.mixin import AdminFiltersMixin
from celery.result import AsyncResult
from import_export import fields, resources
from import_export.admin import ImportExportMixin
from import_export.widgets import ForeignKeyWidget
from smart_admin.mixins import DisplayAllMixin, LinkedObjectsMixin

from power_query.celery_tasks import (
    refresh_report,
    refresh_reports,
    run_background_query,
)
from power_query.forms import FormatterTestForm
from power_query.models import (
    CeleryEnabled,
    Dataset,
    Formatter,
    Parametrizer,
    Query,
    Report,
    ReportDocument,
)
from power_query.utils import to_dataset
from power_query.widget import FormatterEditor, PythonFormatterEditor

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from django.contrib.admin import _ModelT


class CeleryEnabledMixin:
    def get_readonly_fields(
        self, request: HttpRequest, obj: Optional["_ModelT"] = None
    ) -> Sequence[str]:
        ret = list(super().get_readonly_fields(request, obj))
        ret.append("celery_task")
        return ret

    @button(visible=settings.DEBUG)
    def check_status(self, request: HttpRequest) -> HttpResponse:  # type: ignore
        obj: CeleryEnabled
        for obj in self.get_queryset(request):
            if obj.async_result is None:
                obj.celery_task = None
                obj.save()

    @button(visible=settings.DEBUG)
    def monitor(self, request: HttpRequest) -> HttpResponse:
        ctx = self.get_common_context(request, title="Celery Monitor")
        ctx["flower_address"] = settings.FLOWER_ADDRESS
        ctx["queries"] = Query.objects.all()
        ctx["reports"] = Report.objects.all()
        return render(
            request,
            f"admin/power_query/{self.model._meta.model_name}/monitor.html",
            ctx,
        )

    @button()
    def queue(self, request: HttpRequest, pk: int) -> HttpResponse:  # type: ignore
        obj: Optional[Any]
        try:
            obj = self.get_object(request, str(pk))
            obj.queue()
            self.message_user(request, f"Run scheduled: {obj.celery_task}")
        except Exception as e:
            self.message_user(request, f"{e.__class__.__name__}: {e}", messages.ERROR)


@register(Query)
class QueryAdmin(
    AdminFiltersMixin,
    LinkedObjectsMixin,
    CeleryEnabledMixin,
    ExtraButtonsMixin,
    DisplayAllMixin,
    AdminActionPermMixin,
    ModelAdmin,
):
    list_display = ("name", "target", "owner", "active", "success")
    search_fields = ("name",)
    list_filter = (
        ("target", AutoCompleteFilter),
        ("owner", AutoCompleteFilter),
        "active",
        "last_run",
    )
    autocomplete_fields = ("target", "owner")
    readonly_fields = ("sentry_error_id", "error_message", "info")
    change_form_template = None

    def success(self, obj: Query) -> bool:
        return not bool(obj.error_message)

    success.boolean = True

    def formfield_for_dbfield(
        self, db_field: Any, request: HttpRequest, **kwargs: Any
    ) -> Optional[forms.fields.Field]:
        if db_field.name == "code":
            kwargs = {"widget": PythonFormatterEditor}
        elif db_field.name == "description":
            kwargs = {"widget": forms.Textarea(attrs={"rows": 2, "style": "width:80%"})}
        elif db_field.name == "owner":
            kwargs = {"widget": forms.HiddenInput}

        return super().formfield_for_dbfield(db_field, request, **kwargs)

    def has_change_permission(
        self, request: HttpRequest, obj: Optional[Any] = None
    ) -> bool:
        return request.user.is_superuser or bool(obj and obj.owner == request.user)

    @button()
    def datasets(self, request: HttpRequest, pk: int) -> HttpResponseRedirect:  # type: ignore[return]
        obj = self.get_object(request, str(pk))
        try:
            url = reverse("admin:power_query_dataset_changelist")
            return HttpResponseRedirect(f"{url}?query__exact={obj.pk}")
        except Exception as e:  # pragma: no cover
            self.message_user(request, f"{e.__class__.__name__}: {e}", messages.ERROR)

    @button(visible=settings.DEBUG)
    def run(self, request: HttpRequest, pk: int) -> HttpResponse:
        ctx = self.get_common_context(request, pk, title="Run results")
        query = self.get_object(request, str(pk))
        results = query.execute_matrix(persist=True)
        self.message_user(request, "Done", messages.SUCCESS)
        ctx["results"] = results
        return render(request, "admin/power_query/query/run_result.html", ctx)

    @button()
    def queue(self, request: HttpRequest, pk: int) -> HttpResponseRedirect:  # type: ignore[return]
        try:
            res: AsyncResult = run_background_query.delay(pk)
            self.message_user(request, f"Query scheduled: {res}")
        except Exception as e:
            self.message_user(request, f"{e.__class__.__name__}: {e}", messages.ERROR)

    @button()
    def preview(self, request: HttpRequest, pk: int) -> HttpResponse:  # type: ignore[return]
        obj = self.get_object(request, str(pk))
        try:
            context = self.get_common_context(request, pk, title="Results")
            if obj.parametrizer:
                args = obj.parametrizer.get_matrix()[0]
            else:
                args = {}
            ret, extra = obj.run(False, args, use_existing=True)
            context["type"] = type(ret).__name__
            context["raw"] = ret
            context["info"] = extra
            context["title"] = f"Result of {obj.name} ({type(ret).__name__})"
            if isinstance(ret, QuerySet):
                ret = ret[:100]
                context["queryset"] = ret
            elif isinstance(ret, tablib.Dataset):
                context["dataset"] = ret[:100]
            elif isinstance(ret, (dict, list, tuple)):
                context["result"] = ret[:100]
            else:
                self.message_user(
                    request,
                    f"Query does not returns a valid result. It returned {type(ret)}",
                )
            return render(request, "admin/power_query/query/preview.html", context)
        except Exception as e:
            self.message_user(request, f"{e.__class__.__name__}: {e}", messages.ERROR)

    def get_changeform_initial_data(self, request: HttpRequest) -> Dict[str, Any]:
        ct = ContentType.objects.filter(id=request.GET.get("ct", 0)).first()
        return {
            "code": "result=conn.all()",
            "name": ct,
            "target": ct,
            "owner": request.user,
        }


@register(Dataset)
class DatasetAdmin(
    AdminFiltersMixin,
    ExtraButtonsMixin,
    DisplayAllMixin,
    AdminActionPermMixin,
    ModelAdmin,
):
    search_fields = ("query__name",)
    list_display = (
        "query",
        "last_run",
        "dataset_type",
        "target_type",
        "size",
        "arguments",
    )
    list_filter = (
        ("query__target", AutoCompleteFilter),
        ("query", AutoCompleteFilter),
        "last_run",
    )
    change_form_template = None
    readonly_fields = ("last_run", "query", "info")
    date_hierarchy = "last_run"

    def get_queryset(self, request: HttpRequest):  # type: ignore[no-untyped-def]
        return super().get_queryset(request).defer("extra", "value")

    def has_add_permission(self, request: HttpRequest) -> bool:
        return False

    def arguments(self, obj: Any) -> str:
        return obj.info.get("arguments")

    def dataset_type(self, obj: Any) -> str:
        return obj.info.get("type")

    def target_type(self, obj: Any) -> str:
        return obj.query.target

    @button(visible=lambda btn: "change" in btn.context["request"].path)
    def preview(self, request: HttpRequest, pk: int) -> HttpResponse:  # type: ignore[return]
        obj = self.get_object(request, str(pk))
        try:
            context = self.get_common_context(request, pk, title="Results")
            context["dataset"] = to_dataset(obj.data)
            return render(request, "admin/power_query/query/preview.html", context)
        except Exception as e:
            logger.exception(e)
            self.message_user(request, f"{e.__class__.__name__}: {e}", messages.ERROR)


class FormatterResource(resources.ModelResource):
    class Meta:
        model = Report
        fields = ("name", "content_type", "code")
        import_id_fields = ("name",)


@register(Formatter)
class FormatterAdmin(
    ImportExportMixin,
    ExtraButtonsMixin,
    DisplayAllMixin,
    AdminActionPermMixin,
    ModelAdmin,
):
    list_display = ("name", "content_type")
    search_fields = ("name",)
    list_filter = ("content_type",)
    resource_class = FormatterResource
    change_form_template = None

    formfield_overrides = {
        models.TextField: {"widget": FormatterEditor(theme="abcdef")},
    }

    @button(visible=lambda btn: "change" in btn.context["request"].path)
    def test(self, request: HttpRequest, pk: int) -> HttpResponse:
        context = self.get_common_context(request, pk)
        form = FormatterTestForm()
        try:
            if request.method == "POST":
                form = FormatterTestForm(request.POST)
                if form.is_valid():
                    obj: Formatter = context["original"]
                    ctx = {
                        "dataset": form.cleaned_data["query"].datasets.first(),
                        "report": "None",
                    }
                    if obj.content_type == "xls":
                        output = obj.render(ctx)
                        response = HttpResponse(output, content_type=obj.content_type)
                        response[
                            "Content-Disposition"
                        ] = "attachment; filename=Report.xls"
                        return response
                    else:
                        context["results"] = str(obj.render(ctx))
                else:
                    form = FormatterTestForm()
        except Exception as e:
            logger.exception(e)
            self.message_user(request, f"{e.__class__.__name__}: {e}", messages.ERROR)
        context["form"] = form
        return render(request, "admin/power_query/formatter/test.html", context)


class ReportResource(resources.ModelResource):
    query = fields.Field(widget=ForeignKeyWidget(Query, "name"))
    formatter = fields.Field(widget=ForeignKeyWidget(Formatter, "name"))

    class Meta:
        model = Report
        fields = ("name", "query", "formatter")
        import_id_fields = ("name",)


@register(Report)
class ReportAdmin(
    AdminFiltersMixin,
    CeleryEnabledMixin,
    LinkedObjectsMixin,
    ExtraButtonsMixin,
    DisplayAllMixin,
    AdminActionPermMixin,
    ModelAdmin,
):
    list_display = ("name", "formatter", "last_run", "frequence", "owner")
    autocomplete_fields = ("query", "formatter", "owner")
    filter_horizontal = ["limit_access_to"]
    readonly_fields = ("last_run",)
    list_filter = (
        ("owner", AutoCompleteFilter),
        ("query", AutoCompleteFilter),
        ("formatter", AutoCompleteFilter),
        "last_run",
    )
    resource_class = ReportResource
    search_fields = ("name",)

    def has_change_permission(
        self, request: HttpRequest, obj: Optional[Any] = None
    ) -> bool:
        return request.user.is_superuser or bool(obj and obj.owner == request.user)

    def get_changeform_initial_data(self, request: HttpRequest) -> Dict:
        kwargs: Dict = {"owner": request.user}
        if "q" in request.GET:
            q = Query.objects.get(pk=request.GET["q"])
            kwargs["query"] = q
            kwargs["name"] = f"Report for {q.name}"
            kwargs["notify_to"] = [request.user]
        return kwargs

    @button(visible=lambda btn: "change" in btn.context["request"].path)
    def queue(self, request: HttpRequest, pk: int) -> HttpResponseRedirect:  # type: ignore[return]
        obj = self.get_object(request, str(pk))
        try:
            res: AsyncResult = refresh_report.delay(obj.pk)
            self.message_user(request, f"Report scheduled: {res}")
        except Exception as e:
            logger.exception(e)
            self.message_user(request, f"{e.__class__.__name__}: {e}", messages.ERROR)

    @button(visible=lambda btn: "change" in btn.context["request"].path)
    def execute(self, request: HttpRequest, pk: int) -> HttpResponseRedirect:  # type: ignore[return]
        obj = self.get_object(request, str(pk))
        try:
            result = obj.execute(run_query=True)
            errors = [r[1] for r in result if isinstance(r[1], Exception)]
            if len(errors) == 0:
                message_level = messages.SUCCESS
            elif len(errors) == len(result):
                message_level = messages.ERROR
            else:
                message_level = messages.WARNING
            self.message_user(request, f"{result}", message_level)
        except Exception as e:
            logger.exception(e)
            self.message_user(request, f"{e.__class__.__name__}: {e}", messages.ERROR)

    @button(
        visible=lambda btn: btn.context["request"].path.endswith("/power_query/report/")
    )
    def refresh(self, request: HttpRequest) -> HttpResponseRedirect:  # type: ignore[return]
        try:
            refresh_reports.delay()
            self.message_user(request, "Reports refresh queued", messages.SUCCESS)
        except Exception as e:
            self.message_user(request, f"{e.__class__.__name__}: {e}", messages.ERROR)


@register(Parametrizer)
class QueryArgsAdmin(
    LinkedObjectsMixin,
    ExtraButtonsMixin,
    DisplayAllMixin,
    AdminActionPermMixin,
    ModelAdmin,
):
    list_display = ("name", "code", "system")
    list_filter = ("system",)
    search_fields = ("name", "code")
    json_enabled = True

    @button()
    def preview(self, request: HttpRequest, pk: int) -> TemplateResponse:
        context = self.get_common_context(request, pk, title="Execution Plan")
        return TemplateResponse(
            request, "admin/power_query/queryargs/preview.html", context
        )

    @button()
    def refresh(self, request: HttpRequest, pk: int) -> None:
        obj = self.get_object(request, str(pk))
        obj.refresh()


@register(ReportDocument)
class ReportDocumentAdmin(
    AdminFiltersMixin,
    LinkedObjectsMixin,
    ExtraButtonsMixin,
    DisplayAllMixin,
    AdminActionPermMixin,
    ModelAdmin,
):
    list_display = ("title", "content_type", "arguments")
    list_filter = (("report", AutoCompleteFilter),)
    filter_horizontal = ("limit_access_to",)
    readonly_fields = ("arguments", "report", "dataset", "content_type")

    def get_queryset(self, request: HttpRequest) -> QuerySet[ReportDocument]:
        return super().get_queryset(request).defer("output")

    def size(self, obj: ReportDocument) -> int:
        return len(obj.output or "")

    @button()
    def view(self, request: HttpRequest, pk: int) -> HttpResponseRedirect:
        if not (obj := self.get_object(request, str(pk))):
            raise Exception("Report document not found")
        url = reverse("power_query:document", args=[obj.report.pk, pk])
        return HttpResponseRedirect(url)

    def has_add_permission(self, request: HttpRequest) -> bool:
        return False



from django.core.exceptions import ValidationError
from django.utils.deconstruct import deconstructible

from power_query.utils import should_run


@deconstructible
class FrequencyValidator:
    message = "Invalid frequency"
    code = "invalid"

    def __call__(self, value: str) -> None:
        try:
            should_run(value)
        except BaseException:
            raise ValidationError(self.message, code=self.code, params={"value": value})



from typing import Any
from uuid import UUID

from django.contrib.gis.gdal.geometries import Point as GdalPoint
from django.contrib.gis.geos import Point as GeosPoint
from django.core.serializers.json import DjangoJSONEncoder
from django.db import models


class PQJSONEncoder(DjangoJSONEncoder):
    def default(self, o: Any) -> Any:
        if isinstance(o, models.Model):
            return str(o)
        if isinstance(o, UUID):
            return o.hex
        if isinstance(o, (GdalPoint, GeosPoint)):
            return [o.x, o.y]
        return super().default(o)



from django.urls import path

from power_query.views import data, document, report, report_list

app_name = "power_query"

urlpatterns = [
    path("reports/", report_list, name="report_list"),
    path("report/<int:pk>/", report, name="report"),
    path("document/<int:report>/<int:pk>/", document, name="document"),
    path("data/<int:pk>/", data, name="data"),
]



from typing import Any, Dict, Optional, TYPE_CHECKING, Union

from django import forms
from django.contrib.contenttypes.models import ContentType
from django.db.models import Q
from django.templatetags.static import static

if TYPE_CHECKING:
    from django.db.models.fields import _ChoicesCallable


class FormatterEditor(forms.Textarea):
    template_name = "power_query/widgets/codewidget.html"

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        theme = kwargs.pop("theme", "midnight")
        super().__init__(*args, **kwargs)
        self.attrs["class"] = "formatter-editor"
        self.attrs["theme"] = theme

    class Media:
        css = {
            "all": (
                static("admin/power_query/codemirror/codemirror.css"),
                static("admin/power_query/codemirror/fullscreen.css"),
                static("admin/power_query/codemirror/foldgutter.css"),
                static("admin/power_query/codemirror/midnight.css"),
                static("admin/power_query/codemirror/abcdef.css"),
            )
        }
        js = (
            static("admin/power_query/codemirror/codemirror.js"),
            static("admin/power_query/codemirror/fullscreen.js"),
            static("admin/power_query/codemirror/active-line.js"),
            static("admin/power_query/codemirror/foldcode.js"),
            static("admin/power_query/codemirror/foldgutter.js"),
            static("admin/power_query/codemirror/indent-fold.js"),
            static("admin/power_query/codemirror/overlay.js"),
        )


class PythonFormatterEditor(FormatterEditor):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.attrs["class"] = "python-editor"

    class Media(FormatterEditor.Media):
        js = FormatterEditor.Media.js + (
            static("admin/power_query/codemirror/python.js"),
            static("admin/power_query/codemirror/django.js"),
        )


class ContentTypeChoiceField(forms.ModelChoiceField):
    def __init__(
        self,
        *,
        empty_label: str = "---------",
        required: bool = True,
        widget: Optional[Any] = None,
        label: Optional[Any] = None,
        initial: Optional[Any] = None,
        help_text: str = "",
        to_field_name: Optional[str] = None,
        limit_choices_to: Union[
            Union[Q, Dict[str, Any]], "_ChoicesCallable", None
        ] = None,
        **kwargs: Any,
    ):
        queryset = ContentType.objects.order_by("model", "app_label")
        super().__init__(
            queryset,
            empty_label=empty_label,
            required=required,
            widget=widget,
            label=label,
            initial=initial,
            help_text=help_text,
            to_field_name=to_field_name,
            limit_choices_to=limit_choices_to,
            **kwargs,
        )

    def label_from_instance(self, obj: Any) -> str:
        return f"{obj.name.title()} ({obj.app_label})"



from django.test import override_settings, TestCase
from django.urls import reverse

from power_query.defaults import create_defaults
from power_query.fixtures import (
    FormatterFactory,
    ParametrizerFactory,
    QueryFactory,
    ReportFactory,
    UserFactory,
)
from power_query.models import Formatter, Query, Report


@override_settings(POWER_QUERY_DB_ALIAS="default")
class TestPowerQuery(TestCase):
    databases = {"default"}

    @classmethod
    def setUpTestData(cls) -> None:
        cls.superuser = UserFactory(is_superuser=True, is_staff=True, is_active=True)

        # code should be unique but the test depends on both BAs having the same, empty code
        cls.user1 = UserFactory(is_superuser=False, is_staff=False, is_active=True)
        cls.user2 = UserFactory(is_superuser=False, is_staff=False, is_active=True)
        cls.user3 = UserFactory(is_superuser=False, is_staff=False, is_active=True)
        create_defaults()

        p = ParametrizerFactory()

        cls.query1: Query = QueryFactory(
            name="Query1", code="result=conn.all()", parametrizer=p
        )
        res = cls.query1.execute_matrix()
        assert res
        cls.formatter: Formatter = FormatterFactory(name="Queryset To HTML")
        cls.report1: Report = ReportFactory(formatter=cls.formatter, query=cls.query1)
        cls.report1.execute()

    def test_access_forbidden(self) -> None:
        url = reverse(
            "power_query:document",
            args=[self.report1.pk, self.report1.documents.first().pk],
        )
        self.client.login(username=self.user2.username, password="password")
        response = self.client.get(url)
        self.assertEqual(response.status_code, 403)



import os

from celery import Celery

os.environ.setdefault("DJANGO_SETTINGS_MODULE", "demo.settings")

app = Celery("demo")

app.config_from_object("django.conf:settings", namespace="CELERY")
app.autodiscover_tasks(related_name="celery_tasks")



from django import forms
from django.contrib.auth import get_user_model

from power_query.models import Formatter, Query
from power_query.widget import ContentTypeChoiceField, PythonFormatterEditor


class ExportForm(forms.Form):
    formatter = forms.ModelChoiceField(queryset=Formatter.objects)  # type: ignore # Argument "queryset" to "ModelChoiceField" has incompatible type "BaseManager[Any]"; expected "Union[None, Manager[Model], _QuerySet[Model, Model]]"


class FormatterTestForm(forms.Form):
    query = forms.ModelChoiceField(Query.objects)  # type: ignore # Argument 1 to "ModelChoiceField" has incompatible type "BaseManager[Any]"; expected "Union[None, Manager[Model], _QuerySet[Model, Model]]"


class QueryForm(forms.ModelForm):
    name = forms.CharField(
        required=True, widget=forms.TextInput(attrs={"style": "width:80%"})
    )
    target = ContentTypeChoiceField()
    code = forms.CharField(widget=PythonFormatterEditor)
    owner = forms.ModelChoiceField(queryset=get_user_model().objects, widget=forms.HiddenInput)  # type: ignore # Argument "queryset" to "ModelChoiceField" has incompatible type "BaseManager[Any]"; expected "Union[None, Manager[Model], _QuerySet[Model, Model]]"
    description = forms.CharField(
        required=False, widget=forms.Textarea(attrs={"rows": 2, "style": "width:80%"})
    )

    class Meta:
        model = Query
        fields = ("name", "target", "description", "code")



from django.contrib.contenttypes.management import create_contenttypes
from django.test import override_settings
from django.urls import reverse

from django_webtest import WebTest

from power_query.defaults import create_defaults
from power_query.fixtures import (
    DatasetFactory,
    FormatterFactory,
    ParametrizerFactory,
    QueryFactory,
    ReportDocumentFactory,
    ReportFactory,
    UserFactory,
)
from power_query.models import Parametrizer, Query


@override_settings(POWER_QUERY_DB_ALIAS="default")
class TestPowerQueryAdmin(WebTest):
    databases = {"default"}

    @classmethod
    def setUpTestData(cls) -> None:
        cls.superuser = UserFactory(is_superuser=True, is_staff=True, is_active=True)
        cls.user1 = UserFactory(is_superuser=False, is_staff=False, is_active=True)
        cls.user2 = UserFactory(is_superuser=False, is_staff=False, is_active=True)
        create_defaults()

        from django.apps import apps

        create_contenttypes(apps.get_app_config("auth"))
        cls.query_qs = QueryFactory()
        cls.query_values = QueryFactory(
            name="byValue", code="result=list(conn.values())"
        )
        cls.query_error = QueryFactory(name="Error", code="result=1")
        cls.query_dataset = QueryFactory(code="result=Dataset(conn.values())")
        cls.query_params = QueryFactory(parametrizer=ParametrizerFactory())

    def test_change_list(self) -> None:
        url = reverse("admin:power_query_query_changelist")
        res = self.app.get(url, user=self.superuser)
        assert res.status_code == 200

    def test_change_form(self) -> None:
        url = reverse("admin:power_query_query_change", args=[self.query_qs.pk])
        res = self.app.get(url, user=self.superuser)
        res = res.forms["query_form"].submit()
        assert res.status_code == 302

    def test_button_run(self) -> None:
        url = reverse("admin:power_query_query_change", args=[self.query_qs.pk])
        res = self.app.get(url, user=self.superuser)
        res = res.click("Run")
        assert res.status_code == 200

    def test_button_datasets(self) -> None:
        url = reverse("admin:power_query_query_change", args=[self.query_qs.pk])
        res = self.app.get(url, user=self.superuser)
        res = res.click("Datasets", linkid="btn-datasets")
        assert res.status_code == 302

    def test_button_queue(self) -> None:
        url = reverse("admin:power_query_query_change", args=[self.query_qs.pk])
        res = self.app.get(url, user=self.superuser)
        res = res.click("Queue")
        assert res.status_code == 302

    def test_button_preview_qs(self) -> None:
        url = reverse("admin:power_query_query_change", args=[self.query_qs.pk])
        res = self.app.get(url, user=self.superuser)
        res = res.click("Preview")
        assert res.status_code == 200

    def test_button_preview_values(self) -> None:
        url = reverse("admin:power_query_query_change", args=[self.query_values.pk])
        res = self.app.get(url, user=self.superuser)
        res = res.click("Preview")
        assert res.status_code == 200

    def test_button_preview_dataset(self) -> None:
        url = reverse("admin:power_query_query_change", args=[self.query_dataset.pk])
        res = self.app.get(url, user=self.superuser)
        res = res.click("Preview")
        assert res.status_code == 200

    def test_button_preview_value_error(self) -> None:
        url = reverse("admin:power_query_query_change", args=[self.query_error.pk])
        res = self.app.get(url, user=self.superuser)
        res = res.click("Preview")
        assert res.status_code == 200

    def test_button_preview_parametrizer(self) -> None:
        url = reverse("admin:power_query_query_change", args=[self.query_params.pk])
        res = self.app.get(url, user=self.superuser)
        res = res.click("Preview")
        assert res.status_code == 200

    def test_create(self) -> None:
        url = reverse("admin:power_query_query_add")
        res = self.app.get(url, user=self.superuser)
        assert res.status_code == 200


@override_settings(POWER_QUERY_DB_ALIAS="default")
class TestDatasetAdmin(WebTest):
    databases = {"default"}

    @classmethod
    def setUpTestData(cls) -> None:
        cls.superuser = UserFactory(is_superuser=True, is_staff=True, is_active=True)
        cls.user1 = UserFactory(is_superuser=False, is_staff=False, is_active=True)
        cls.user2 = UserFactory(is_superuser=False, is_staff=False, is_active=True)
        create_defaults()
        cls.dataset = DatasetFactory()

    def test_change_list(self) -> None:
        url = reverse("admin:power_query_dataset_changelist")
        res = self.app.get(url, user=self.superuser)
        assert res.status_code == 200

    def test_change_form(self) -> None:
        url = reverse("admin:power_query_dataset_change", args=[self.dataset.pk])
        res = self.app.get(url, user=self.superuser)
        assert res.status_code == 200

    def test_button_preview(self) -> None:
        url = reverse("admin:power_query_dataset_change", args=[self.dataset.pk])
        res = self.app.get(url, user=self.superuser)
        res = res.click("Preview")
        assert res.status_code == 200


@override_settings(POWER_QUERY_DB_ALIAS="default")
class TestFormatterAdmin(WebTest):
    databases = {"default"}

    @classmethod
    def setUpTestData(cls) -> None:
        cls.superuser = UserFactory(is_superuser=True, is_staff=True, is_active=True)
        cls.user1 = UserFactory(is_superuser=False, is_staff=False, is_active=True)
        cls.user2 = UserFactory(is_superuser=False, is_staff=False, is_active=True)
        create_defaults()
        cls.formatter = FormatterFactory()
        QueryFactory()

    def test_change_list(self) -> None:
        url = reverse("admin:power_query_formatter_changelist")
        res = self.app.get(url, user=self.superuser)
        assert res.status_code == 200

    def test_change_form(self) -> None:
        url = reverse("admin:power_query_formatter_change", args=[self.formatter.pk])
        res = self.app.get(url, user=self.superuser)
        assert res.status_code == 200

    def test_button_test(self) -> None:
        url = reverse("admin:power_query_formatter_change", args=[self.formatter.pk])
        res = self.app.get(url, user=self.superuser)
        res = res.click("Test")
        res.forms["formatter-form"]["query"] = Query.objects.first().pk
        res = res.forms["formatter-form"].submit()
        assert res.status_code == 200


@override_settings(POWER_QUERY_DB_ALIAS="default")
class TestReportAdmin(WebTest):
    databases = {"default"}

    @classmethod
    def setUpTestData(cls) -> None:
        cls.superuser = UserFactory(is_superuser=True, is_staff=True, is_active=True)
        cls.user1 = UserFactory(is_superuser=False, is_staff=False, is_active=True)
        cls.user2 = UserFactory(is_superuser=False, is_staff=False, is_active=True)
        create_defaults()
        QueryFactory()
        cls.report = ReportFactory()

    def test_change_list(self) -> None:
        url = reverse("admin:power_query_report_changelist")
        res = self.app.get(url, user=self.superuser)
        assert res.status_code == 200

    def test_add(self) -> None:
        url = reverse("admin:power_query_report_add")
        res = self.app.get(url, user=self.superuser)
        assert res.status_code == 200

    def test_add_param(self) -> None:
        url = reverse("admin:power_query_report_add")
        qid = QueryFactory().pk
        res = self.app.get(f"{url}?q={qid}", user=self.superuser)
        assert res.status_code == 200

    def test_change_form(self) -> None:
        url = reverse("admin:power_query_report_change", args=[self.report.pk])
        res = self.app.get(url, user=self.superuser)
        assert res.status_code == 200

    def test_button_queue(self) -> None:
        url = reverse("admin:power_query_report_change", args=[self.report.pk])
        res = self.app.get(url, user=self.superuser)
        res = res.click("Queue")
        assert res.status_code == 302

    def test_button_execute(self) -> None:
        url = reverse("admin:power_query_report_change", args=[self.report.pk])
        res = self.app.get(url, user=self.superuser)
        res = res.click("Execute")
        assert res.status_code == 302

    def test_button_refresh(self) -> None:
        url = reverse("admin:power_query_report_changelist")
        res = self.app.get(url, user=self.superuser)
        res = res.click("Refresh")
        assert res.status_code == 302


@override_settings(POWER_QUERY_DB_ALIAS="default")
class TestParametrizerAdmin(WebTest):
    databases = {"default"}

    @classmethod
    def setUpTestData(cls) -> None:
        cls.superuser = UserFactory(is_superuser=True, is_staff=True, is_active=True)
        cls.user1 = UserFactory(is_superuser=False, is_staff=False, is_active=True)
        cls.user2 = UserFactory(is_superuser=False, is_staff=False, is_active=True)
        create_defaults()
        cls.params = ParametrizerFactory()
        cls.system_params = Parametrizer.objects.get(code="active-business-areas")

    def test_button_refresh(self) -> None:
        url = reverse("admin:power_query_parametrizer_change", args=[self.system_params.pk])
        res = self.app.get(url, user=self.superuser)
        res = res.click("Refresh")
        assert res.status_code == 302

    def test_button_preview(self) -> None:
        url = reverse("admin:power_query_parametrizer_change", args=[self.params.pk])
        res = self.app.get(url, user=self.superuser)
        res = res.click("Preview")
        assert res.status_code == 200


@override_settings(POWER_QUERY_DB_ALIAS="default")
class TestReportDocumentAdmin(WebTest):
    databases = {"default"}

    @classmethod
    def setUpTestData(cls) -> None:
        cls.superuser = UserFactory(is_superuser=True, is_staff=True, is_active=True)
        cls.user1 = UserFactory(is_superuser=False, is_staff=False, is_active=True)
        cls.user2 = UserFactory(is_superuser=False, is_staff=False, is_active=True)
        create_defaults()
        cls.document = ReportDocumentFactory()

    def test_change_list(self) -> None:
        url = reverse("admin:power_query_reportdocument_changelist")
        res = self.app.get(url, user=self.superuser)
        assert res.status_code == 200

    def test_change_form(self) -> None:
        url = reverse(
            "admin:power_query_reportdocument_change", args=[self.document.pk]
        )
        res = self.app.get(url, user=self.superuser)
        assert res.status_code == 200



from django.contrib.contenttypes.models import ContentType
from django.http import HttpRequest, HttpResponse
from django.shortcuts import render

from admin_extra_buttons.decorators import button
from admin_extra_buttons.mixins import ExtraButtonsMixin

from power_query.models import Query


class PowerQueryMixin(ExtraButtonsMixin):
    @button(permission="power_query_query_add")
    def power_query(self, request: HttpRequest) -> HttpResponse:
        ct = ContentType.objects.get_for_model(self.model)
        context = self.get_common_context(
            request,
            title="Power Queries",
            ct=ct,
            entries=Query.objects.filter(target=ct),
        )
        return render(request, "power_query/list.html", context)



from typing import Any


class QueryRunError(Exception):
    def __init__(
        self,
        exception: Exception,
        sentry_error_id: Any = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        self.exception = exception
        self.sentry_error_id = sentry_error_id

    def __str__(self) -> str:
        return str(f"{self.exception.__class__.__name__}: {self.exception}")



import base64
import hashlib
import inspect
import json
from datetime import datetime
from functools import wraps
from typing import Any, Callable, Dict

from django.conf import settings
from django.contrib.auth import authenticate
from django.db.models import QuerySet
from django.http import HttpRequest, HttpResponse
from django.utils.safestring import mark_safe

import tablib
from concurrency.utils import get_classname
from sentry_sdk import configure_scope


def fqn(o: Any) -> str:
    parts = []

    if inspect.isclass(o):
        cls = o
    else:
        cls = type(o)
    if hasattr(o, "__module__"):
        parts.append(o.__module__)
        parts.append(get_classname(o))
    elif inspect.ismodule(o):
        return o.__name__
    else:
        parts.append(cls.__name__)
    if not parts:
        raise ValueError("Invalid argument `{}`".format(o))
    return ".".join(parts)


def to_dataset(result: Any) -> tablib.Dataset:
    if isinstance(result, QuerySet):
        data = tablib.Dataset()
        fields = result.__dict__["_fields"]
        if not fields:
            fields = [field.name for field in result.model._meta.concrete_fields]
        data.headers = fields
        try:
            for obj in result.all():
                data.append(
                    [
                        obj[f] if isinstance(obj, dict) else str(getattr(obj, f))
                        for f in fields
                    ]
                )
        except Exception:
            raise ValueError("Results can't be rendered as a tablib Dataset")
    elif isinstance(result, (list, tuple)):
        data = tablib.Dataset()
        fields = set().union(*(d.keys() for d in list(result)))
        data.headers = fields
        try:
            for obj in result:
                data.append([obj[f] for f in fields])
        except Exception:
            raise ValueError("Results can't be rendered as a tablib Dataset")
    elif isinstance(result, (tablib.Dataset, dict)):
        data = result
    else:
        raise ValueError(f"{result} ({type(result)}")
    return data


def get_sentry_url(event_id: int, html: bool = False) -> str:
    url = f"{settings.SENTRY_URL}?query={event_id}"
    if html:
        return mark_safe('<a href="{url}" target="_sentry" >View on Sentry<a/>')
    return url


def basicauth(view: Callable) -> Callable:
    def wrap(request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:
        if request.user.is_authenticated:
            return view(request, *args, **kwargs)

        if "HTTP_AUTHORIZATION" in request.META:
            auth = request.headers["Authorization"].split()
            if len(auth) == 2:
                if auth[0].lower() == "basic":
                    uname, passwd = (
                        base64.b64decode(auth[1].encode()).decode().split(":")
                    )
                    user = authenticate(username=uname, password=passwd)
                    if user is not None and user.is_active:
                        request.user = user
                        return view(request, *args, **kwargs)

        response = HttpResponse()
        response.status_code = 401
        response["WWW-Authenticate"] = 'Basic realm="HOPE"'
        return response

    return wrap


def sizeof(num: float, suffix: str = "") -> str:
    for unit in ["&nbsp;&nbsp;", "Kb", "Mb", "Gb", "Tb", "Pb", "Eb", "Zb"]:
        if abs(num) < 1024.0:
            return f"{num:3.1f} {unit}{suffix} "
        num /= 1024.0
    return f"{num:.1f}Yi{suffix}"


def dict_hash(dictionary: Dict[str, Any]) -> str:
    """MD5 hash of a dictionary."""
    dhash = hashlib.md5()
    # We need to sort arguments so {'a': 1, 'b': 2} is
    # the same as {'b': 2, 'a': 1}
    encoded = json.dumps(dictionary, sort_keys=True).encode()
    dhash.update(encoded)
    return dhash.hexdigest()


def should_run(expression: str) -> bool:
    match_expressions = expression.split(",")
    today = datetime.today()

    for exp in match_expressions:
        if exp.lower() in ["mon", "tue", "wed", "thu", "fri", "sat", "sun"]:
            if exp.lower() == datetime.today().strftime("%a").lower():
                return True
        elif exp.isnumeric():
            if today.day == int(exp):
                return True
        elif exp.count("/") == 1:
            day, month = exp.split("/")
            if (
                day.isnumeric()
                and month.isnumeric()
                and int(day) == today.day
                and int(month) == today.month
            ):
                return True
        else:
            raise ValueError(f"Invalid expression '{expression}'")
    return False


def sentry_tags(func: Callable) -> Callable:
    """
    add sentry tags 'celery' and 'celery_task'
    """

    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        with configure_scope() as scope:
            scope.set_tag("celery", True)
            scope.set_tag("celery_task", func.__name__)

            return func(*args, **kwargs)

    return wrapper
